package work.xuye.common.spel;

import com.google.gson.JsonArray;
import com.google.gson.JsonObject;
import com.google.gson.JsonParser;
import com.google.gson.JsonPrimitive;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.expression.Expression;
import org.springframework.expression.spel.standard.SpelExpressionParser;
import org.springframework.stereotype.Component;
import org.springframework.util.DigestUtils;
import org.springframework.util.MultiValueMap;
import org.springframework.util.ObjectUtils;
import org.springframework.web.util.UriComponentsBuilder;
import work.xuye.common.constant.CustomerDateTimeFormatter;
import work.xuye.common.constant.RawDataFiledKey;
import work.xuye.common.store.NsKVMapStore;
import work.xuye.common.utils.HttpUtil;
import work.xuye.common.utils.JsonPathUtil;

import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Arrays;
import java.util.Map;

/**
 * @author xuye
 * @since 2023/3/8 12:36
 **/
@Slf4j
@Component
@RequiredArgsConstructor
public class CustomFunction {
    private final JsonPathHandler jsonPathHandler;

    private final SpelExpressionParser parser;

    private final NsKVMapStore mapStore;

    public String jsonExtract(String json, String path) {
        String result = null;
        try {
            result = jsonPathHandler.extract(json, path);
        } catch (Exception e) {
            log.error("jsonExtra handler error, json: [{}], path: [{}]", json, path);
            throw new RuntimeException(e);
        }
        log.debug("jsonExtra handler, json: [{}], path: [{}], result: [{}]", json, path, result);
        return result;
    }

    public String jsonExtractReplaceAll(String json, String jsonPath, String regex, String replacement) {
        return this.jsonExtract(json, jsonPath).replaceAll(regex, replacement).trim();
    }

    public String jsonExtractAppend(String json, String path, String append) {
        String before = this.jsonExtract(json, path);
        String result = before + append;
        log.debug("append handler, before: [{}], after: [{}]", before, result);
        return result;
    }

    public String rfc1123(String json, String path) {
        String before = this.jsonExtract(json, path);
        String result = null;
        try {
            result = ZonedDateTime
                    .parse(before, DateTimeFormatter.RFC_1123_DATE_TIME)
                    .toLocalDateTime()
                    .format(CustomerDateTimeFormatter.READABLE_FORMATTER);
        } catch (Exception e) {
            log.error("rfc1123 handler error, before: [{}]", before);
            throw new RuntimeException(e);
        }
        log.debug("rfc1123 handler, before: [{}], after: [{}]", before, result);
        return result;
    }

    //https://deepq.yuque.com/deepq/bur3gk/hgcvin90uvgyaqmk#FUENY
    public String jnzTitle(String json, String titlePath, String descPath) {
        String result = this.jsonExtract(json, titlePath);
        if (!result.equals("快讯")) {
            log.debug("jnzTitle handler, title: [{}]", result);
            return result;
        } else {
            String desc = this.jsonExtract(json, descPath);
            String descTitle = desc.substring(desc.indexOf("【") + 1, desc.indexOf("】"));
            log.debug("jnzTitle handler, title from desc: [{}]", descTitle);
            return descTitle;
        }
    }

    //https://deepq.yuque.com/deepq/bur3gk/hgcvin90uvgyaqmk#FUENY
    public String jnzDesc(String json, String descPath) {
        String desc = this.jsonExtract(json, descPath);
        if (desc.startsWith("【") && desc.contains("】")) {
            desc = desc.substring(desc.indexOf("】") + 1);
            log.debug("jnzDesc handler, desc after remove: [{}]", desc);
            return desc;
        } else {
            log.debug("jnzDesc handler, desc: [{}]", desc);
            return desc;
        }
    }

    /**
     * 这个方法是为了兼容一些特殊的情况，比如：金牛座的快讯没有link，会导致无法入库，因此使用desc的md5值作为link
     * 优先使用primaryPath，如果primaryPath为空，则使用secondaryPathAsMd5的值
     *
     * @param json               json字符串
     * @param primaryPath        优先使用的路径
     * @param secondaryPathAsMd5 备选路径
     * @return 提取结果
     */
    public String orMd5(String json, String primaryPath, String secondaryPathAsMd5) {
        String primary = this.jsonExtract(json, primaryPath);
        if (!ObjectUtils.isEmpty(primary)) {
            log.debug("orMd5 handler, primary: [{}]", primary);
            return primary;
        } else {
            String sec = this.jsonExtract(json, secondaryPathAsMd5);
            String md5DigestAsHex = DigestUtils.md5DigestAsHex(sec.getBytes());
            log.debug("orMd5 handler, secondary: [{}], md5: [{}]", sec, md5DigestAsHex);
            return md5DigestAsHex;
        }
    }

    /**
     * !!需要注意使用了此方法的任务不要有更新逻辑！！
     * !!需要注意使用了此方法的任务不要有更新逻辑！！
     * !!需要注意使用了此方法的任务不要有更新逻辑！！
     * <p>
     * 作用是获得系统当前的时间，格式化为yyyy-MM-dd HH:mm:ss
     *
     * @return 当前时间
     */
    public String now() {
        Expression exp = parser.parseExpression("T(java.time.LocalDateTime).now().format(T(java.time.format.DateTimeFormatter).ofPattern('yyyy-MM-dd HH:mm:ss'))");
        return (String) exp.getValue();
    }

    public String warpExclude(String json, String excludeFiled, String... fields) {
        JsonObject obj = JsonParser.parseString(json).getAsJsonObject();
        if (obj.has(excludeFiled)) {
            obj.remove(excludeFiled);
        } else {
            log.warn("json warpV2 excludeFiled not exist, excludeFiled: [{}]", excludeFiled);
        }
        return this.wrap(obj.toString(), fields);
    }

    /**
     * 用itemData包一层的原因：避免下游程序直接读到了有特殊意义的字段，比如type(财联社资讯类型 -1：快讯，0：要闻，1：非财联社资讯)
     * <a href="https://app.asana.com/0/1202103682792595/1204211768982825/f">...</a>
     *
     * @param json   json字符串
     * @param fields 需要包装的字段，格式为：key1, path1, key2, path2, ...
     * @return 包装后的json
     */
    public String wrap(String json, String... fields) {
        JsonObject result = new JsonObject();
        try {
            JsonObject obj = JsonParser.parseString(json).getAsJsonObject();
            // 默认使用item包裹一层
            result.add(RawDataFiledKey.ITEM, obj);
            // 默认添加app字段，便于后续区分是不是本程序处理过的数据
            result.add(RawDataFiledKey.CRAWLER_KEY, new JsonPrimitive(RawDataFiledKey.CRAWLER_VALUE));
            if (fields.length % 2 != 0) {
                throw new RuntimeException("fields length must be even! fields: " + Arrays.toString(fields));
            }
            if (fields.length > 0) {
                for (int i = 0; i < fields.length; i += 2) {
                    String key = fields[i];
                    String path = fields[i + 1];
                    String value;
                    if (JsonPathUtil.isJsonPath(path)) {
                        value = this.jsonExtract(json, path);
                    } else {
                        value = path;
                    }
                    result.add(key, new JsonPrimitive(value));
                    log.debug("wrap handler, json: [{}], key: [{}], path: [{}], value: [{}]", json, key, path, value);
                }
            }
            return result.toString();
        } catch (Exception e) {
            log.error("wrap error, json: [{}], fields: [{}]", json, Arrays.toString(fields));
            throw new RuntimeException(e);
        }
    }

    public boolean isHttpRes2XX(String url) {
        return HttpUtil.isHttpRes2XX(url);
    }

    public boolean urlNotContains(String url, String contains) {
        return !url.contains(contains);
    }

    /**
     * 根据key从命名空间中获取value
     *
     * @param namespace 任务名
     * @param key       key
     * @return value
     */
    public String getValueByKeyFromNs(String namespace, String key) {
        String value = mapStore.getValue(namespace, key);
        log.debug("getValueByKeyFromNs handler, namespace: [{}], key-in: [{}], value-out: [{}]", namespace, key, value);
        return value;
    }

    public String getValueByKeyFromNs(String namespace, String json, String jsonPath) {
        String key = this.jsonExtract(json, jsonPath);
        return getValueByKeyFromNs(namespace, key);
    }


    public void putValueToNs(String namespace, String key, String value) {
        mapStore.upsert(namespace, key, value);
        log.debug("putValueToNs handler, namespace: [{}], key: [{}], value: [{}]", namespace, key, value);
    }

    public void putValueToNs(String namespace, String json, String jsonPath, String value) {
        String key = this.jsonExtract(json, jsonPath);
        putValueToNs(namespace, key, value);
    }

    public String getParamsFromUrl(String json, String jsonPath, String key) {
        String url = this.jsonExtract(json, jsonPath);
        return getParamsFromUrl(url, key);
    }

    public String getParamsFromUrl(String url, String key) {
        UriComponentsBuilder builder = UriComponentsBuilder.fromHttpUrl(url);
        MultiValueMap<String, String> queryParams = builder.build().getQueryParams();
        String value = queryParams.getFirst(key);
        log.debug("getParamsFromUrl handler, url: [{}], key-in: [{}], value-out: [{}]", url, key, value);
        return value;
    }


    /**
     * 先从map中获取指定的key，然后使用jsonpath提取json中的数据，判断是否等于期望值
     *
     * @param map         map
     * @param mapFieldKey map中的key
     * @param jsonPath    jsonpath
     * @param value       期望值
     * @return 是否相等
     */
    public boolean equals(Map<String, Object> map, String mapFieldKey, String jsonPath, Object value) {
        Object fieldValue = map.get(mapFieldKey);
        log.debug("equals handler, map: [{}], mapFieldKey: [{}], mapValue: [{}]", map, mapFieldKey, value);
        String result = this.jsonExtract(fieldValue.toString(), jsonPath);
        boolean equals = result.equals(value);
        log.debug("equals handler, json: [{}], jsonPath: [{}], value: [{}], result: [{}]", fieldValue, jsonPath, value, equals);
        return equals;
    }

    public boolean keyNotExistInNs(Map<String, Object> map, String field, String jsonPath, String namespace) {
        return !keyExistInNs(map, field, jsonPath, namespace);
    }

    public boolean keyExistInNs(Map<String, Object> map, String field, String jsonPath, String namespace) {
        Object o = map.get(field);
        String v = this.jsonExtract(o.toString(), jsonPath);
        String queryResult = mapStore.getValue(namespace, v);
        boolean has = !ObjectUtils.isEmpty(queryResult);
        log.info("keyExistInNs handler, value: [{}], namespace: [{}], result: [{}]", v, namespace, has);
        return has;
    }


    /**
     * 固定返回 true，用于希望断言通过时使用
     *
     * @return true
     */
    public boolean returnTrue() {
        return true;
    }


    /**
     * 固定返回 false，用于希望断言不通过时使用
     *
     * @return false
     */
    public boolean returnFalse() {
        return false;
    }

    /**
     * 原始栏目code递增映射
     *
     * @param json     json字符串
     * @param jsonPath jsonpath
     * @return 原始栏目JSON结果
     */
    public synchronized String mappingColumns(String json, String jsonPath) {
        int start = 0;
        String columns = this.jsonExtract(json, jsonPath);
        String[] columnList = columns.split(",");
        JsonArray array = new JsonArray();
        for (String column : columnList) {
            JsonObject obj = new JsonObject();
            obj.addProperty("code", start++);
            obj.addProperty("name", column);
            array.add(obj);
        }
        log.info("incrColumns handler, jsonPath: [{}], columns: [{}], result: [{}]", jsonPath, columns, array);
        return array.toString();
    }

    public String epochMilliToLocalDatetime(String json, String path) {
        String epochMilli = this.jsonExtract(json, path);
        String localDatetime = LocalDateTime.ofInstant(Instant.ofEpochMilli(Long.parseLong(epochMilli)), ZoneId.of("Asia/Shanghai")).format(DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss"));
        log.debug("epochMilliToLocalDatetime handler, json: [{}], path: [{}], epochMilli: [{}], localDatetime: [{}]", json, path, epochMilli, localDatetime);
        return localDatetime;
    }


    public String ifEqualsIgnoreCase(String json, String jsonpath, String dest, String ifValue, String elseValue) {
        String value = this.jsonExtract(json, jsonpath);
        if (value.equalsIgnoreCase(dest)) {
            return ifValue;
        } else {
            return elseValue;
        }
    }


}
