package work.xuye.sink.handler;

import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.messaging.Message;
import org.springframework.stereotype.Component;
import org.springframework.util.ObjectUtils;
import work.xuye.common.constant.MessageConstants;
import work.xuye.common.constant.RegexConstants;
import work.xuye.common.db.entity.Sql;
import work.xuye.common.db.entity.Task;
import work.xuye.common.db.entity.vo.SinkParams;
import work.xuye.common.db.entity.vo.TableTemplate;
import work.xuye.common.db.service.SqlManager;
import work.xuye.common.db.service.TaskManager;
import work.xuye.common.dto.TaskVO;
import work.xuye.common.enums.DeleteCheckMode;
import work.xuye.common.enums.MessageType;
import work.xuye.common.enums.Status;
import work.xuye.common.properties.AlertProperties;
import work.xuye.common.service.MessageService;
import work.xuye.common.spel.SpringExpressionLanguageEvaluator;
import work.xuye.common.utils.DebugUtil;
import work.xuye.sink.service.JdbcService;
import work.xuye.sink.service.SqlExecutor;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.stream.Collectors;

/**
 * @author xuye
 * @since 2023/3/14 09:31
 **/
@Slf4j
@Component
@RequiredArgsConstructor
public class DisappearHandler {
    private final TaskManager taskManager;
    private final JdbcService jdbcService;
    private final MessageService messageService;
    private final SqlManager sqlManager;
    private final AlertProperties alertProperties;
    private final SqlExecutor sqlExecutor;
    private final SpringExpressionLanguageEvaluator evaluator;

    public void handle(Message<Set<String>> message) {


        Integer taskId = (Integer) message.getHeaders().get(MessageConstants.TASK_ID);
        TaskVO taskVO = taskManager.getTaskInfoByTaskId(taskId);
        if (DebugUtil.isSkip(taskVO)) {
            return;
        }
        Task task = taskVO.getTask();
        String taskName = task.getName();
        SinkParams.CheckDelete delete = task.getSinkParams().getCheckDelete();
        if (!delete.getMode().equals(DeleteCheckMode.disappear)) {
            return;
        }
        if (delete.getStatus().equals(Status.off)) {
            return;
        }
        String condition = delete.getDisappearConfig().getCondition();

        String templateNameExpression = taskVO.getTask().getSinkParams().getTemplateNameExpression();
        String templateName = evaluator.evaluate(message.getPayload(), templateNameExpression, String.class).toString();
        TableTemplate template = taskManager.getTemplateByName(templateName).getTableTemplate();
        String tableName = template.getTableName();
        String dataSourceName = template.getDatasourceName();
        List<Map<String, Object>> itemList = jdbcService.getForListMap(dataSourceName,
                SqlGenerator.generateGivenFieldSql(tableName, condition, taskName));
        String uniqueField = template.getUniqueField().getFieldName();
        List<Map<String, Object>> removed = itemList.stream().filter(item -> !message.getPayload().contains(item.get(uniqueField).toString())).collect(Collectors.toList());
        if (ObjectUtils.isEmpty(removed)) {
            return;
        }

        List<String> uniqueFields = removed.stream().map(item -> item.get(uniqueField).toString()).collect(Collectors.toList());
        Sql sqlEntity = sqlManager.saveSql(dataSourceName, SqlGenerator.generateDeleteSqls(tableName, uniqueField, uniqueFields));

        String messageTitle = "";
        String subTitle = "";
        // 是否已经执行过自动删除了
        boolean autoDeleted = false;
        SinkParams.DisappearConfig disappearConfig = task.getSinkParams().getCheckDelete().getDisappearConfig();

        // 处理自动/手动删除，生成通知标题
        if (disappearConfig.isAutoDeleteEnabled() && disappearConfig.getAutoDeleteThreshold() >= removed.size()) {
            sqlExecutor.executeSql(sqlEntity.getId(), sqlEntity.getCaptcha());
            autoDeleted = true;
            messageTitle = "已经自动删除 " + taskName + " 已删除的资讯";
        } else {
            messageTitle = "检测到 " + taskName + " 已删除的资讯，条数超过自动删除阈值，未执行删除，请手动处理";
        }

        // 根据模板，生成通知内容
        String messageTemplate = disappearConfig.getMessageTemplate();
        StringBuffer fullMessage = new StringBuffer();
        Matcher matcher = RegexConstants.PLACEHOLDER_PATTERN.matcher(messageTemplate);
        Map<String, Object> placeholderMap = new HashMap<>();
        while (matcher.find()) {
            placeholderMap.put(matcher.group(), "unknown");
        }
        for (Map<String, Object> item : removed) {
            String messageItem = messageTemplate;
            for (String placeholder : placeholderMap.keySet()) {
                String key = placeholder.replaceAll(RegexConstants.SPECIAL_CHAR_REGEX, "");
                placeholderMap.put(placeholder, item.get(key));
            }
            for (String key : placeholderMap.keySet()) {
                messageItem = messageItem.replace(key, placeholderMap.get(key).toString());
            }
            fullMessage.append(messageItem).append("\r\n\r\n");
        }

        String url = alertProperties.getDomainName().getSink() + "/jdbc/execute/" + sqlEntity.getId() + "/" + sqlEntity.getCaptcha();
        // 如果已经自动删除了，就发送自动删除结果
        subTitle = "待手动删除条数: " + removed.size();
        if (autoDeleted) {
            subTitle = "已自动删除条数: " + removed.size();
            fullMessage.append("> ✅[查看删除结果](").append(url).append(")");
        }
        // 如果未自动删除，并且还配置开启了手动删除的入口，就拼接手动删除的入口
        else if (disappearConfig.isSendExecuteSQLUrl()) {
            fullMessage.append("> ⚠️[点击确认删除](").append(url).append(")");
        }


        // 如果配置了发送通知，就发送通知, 否则就打印日志
        if (disappearConfig.isNotificationEnabled()) {
            messageService.sendMessage(MessageType.info, messageTitle, subTitle, fullMessage.toString());
        } else {
            log.warn("检测到 {} 已删除的资讯, 待删除条数: {}, 未发送通知: {}", taskName, removed.size(), fullMessage);
        }

    }
}
