package work.xuye.sink.handler;

import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.messaging.Message;
import org.springframework.messaging.MessageHeaders;
import org.springframework.stereotype.Component;
import org.springframework.util.Assert;
import org.springframework.util.DigestUtils;
import org.springframework.util.ObjectUtils;
import work.xuye.common.constant.MessageConstants;
import work.xuye.common.constant.StageConstants;
import work.xuye.common.db.entity.Task;
import work.xuye.common.db.entity.vo.SinkParams;
import work.xuye.common.db.entity.vo.TableTemplate;
import work.xuye.common.db.service.TaskManager;
import work.xuye.common.dto.TaskVO;
import work.xuye.common.enums.*;
import work.xuye.common.service.UrlMD5Service;
import work.xuye.common.spel.CustomFunction;
import work.xuye.common.spel.SpringExpressionLanguageEvaluator;
import work.xuye.common.utils.DebugUtil;
import work.xuye.sink.service.JdbcService;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * @author xuye
 * @since 2023/2/19 17:20
 **/
@Slf4j
@Component
@RequiredArgsConstructor
public class ItemHandler {
    private final JdbcService jdbcService;
    private final UrlMD5Service urlMD5Service;
    private final SpringExpressionLanguageEvaluator evaluator;
    private final CustomFunction cf;

    private final TaskManager taskManager;

    private boolean needHandle(Message<String> message) {
        MessageHeaders headers = message.getHeaders();
        String currentStage = (String) headers.get(MessageConstants.CURRENT_STAGE);
        if (StageConstants.END.equals(currentStage)) {
            log.debug("@@@ 此消息已经处于结束状态，无需处理");
            return false;
        }
        return true;
    }


    public void handle(Message<String> message) {
        try {
            this.doHandle(message);
        } catch (Exception e) {
            // 如果这一步出错，需要清除seedUrl缓存，避免上一步的缓存仍然存在，影响执行到本类的下一次重试
            this.cleanCache(message);
            throw new RuntimeException(e);
        }
    }

    private void cleanCache(Message<String> message) {
        String seedUrl = (String) message.getHeaders().get(MessageConstants.SEED_URL);
        urlMD5Service.removeUrlMD5(seedUrl);
    }


    public void doHandle(Message<String> message) {
        if (!needHandle(message)) {
            return;
        }
        TaskVO taskVO = taskManager.getTaskInfoByTaskId((Integer) message.getHeaders().get(MessageConstants.TASK_ID));
        if (DebugUtil.isSkip(taskVO)) {
            return;
        }
        //相关字段放这里
        String md5Digest = DigestUtils.md5DigestAsHex(message.getPayload().getBytes());


        String templateNameExpression = taskVO.getTask().getSinkParams().getTemplateNameExpression();
        String templateName = evaluator.evaluate(message.getPayload(), templateNameExpression, String.class).toString();
        TableTemplate template = taskManager.getTemplateByName(templateName).getTableTemplate();

        SinkParams.CheckDelete checkDeleteConfig = taskVO.getTask().getSinkParams().getCheckDelete();
        // 根据消息体，和映射关系，使用SpEL表达式，得到计算结果
        HashMap<String, Object> expressionResultMap = evaluator.evaluate(taskVO.getTemplateMappingMap().get(templateName).getTableMapping(), message);
        // 如果不能通过校验，则抛出异常
        this.validate(expressionResultMap, template);
        // 将计算结果合并到模板中，此刻的模板是拥有值的
        this.mergeValueToTemplate(taskVO, template, expressionResultMap);
        // 根据MD5判断是否需要处理
        ProcessReason reason = this.getReason(taskVO, template, md5Digest);
        if (reason == ProcessReason.unchanged) {
            // todo 考虑是否有必要加日志
            return;
        }
        // 校验模板中各个字段的表达式
        if (!this.isTemplateValid(taskVO.getTask().getName(), template)) {
            log.info("[{}][\uD83D\uDDD1️ 丢弃数据][{} {}] " + "{}: {}",
                    taskVO.getTask().getName(),
                    reason.getEmoji(),
                    reason.name(),
                    template.getUniqueField().getFieldName(),
                    template.getUniqueField().getValue());
            // 校验不通过，存缓存，避免还处理到这条数据
            urlMD5Service.put(template.getUniqueField().getValue().toString(), md5Digest);
            return;
        }
        //判断数据是否存在
        boolean itemExist = this.itemExist(taskVO.getTask().getSinkParams(), template);
        //如果已经存在了，就判断是否需要删除或更新
        if (itemExist) {
            // 先去尝试删除
            boolean deleted = this.tryDelete(taskVO, template, expressionResultMap, checkDeleteConfig, reason);
            // 如果将已存在的数据删除了，就可以直接放入缓存然后结束了
            if (deleted) {
                this.putToCache(template, md5Digest);
                return;
            }
            Map<String, Object> dbItem = this.getExistedItem(template);
            this.tryUpdate(taskVO, dbItem, expressionResultMap, md5Digest, template, reason);
            //如果不存在，就插入
        } else {
            this.tryInsert(message.getPayload(), taskVO, expressionResultMap, template, reason);
        }
        this.putToCache(template, md5Digest);
    }

    /**
     * 是否满足一组断言规则
     *
     * @param task       任务
     * @param map        map结构的数据
     * @param template   模板
     * @param predicates 断言规则
     * @return 是否满足
     */
    private boolean isSatisfied(Task task, Map<String, Object> map, TableTemplate template, List<String> predicates, String description) {
        if (ObjectUtils.isEmpty(predicates)) {
            return false;
        }
        boolean isValid = true;
        TableTemplate.Field uniqueField = template.getUniqueField();
        String taskName = task.getName();
        for (String expr : predicates) {
            boolean valid = (Boolean) evaluator.evaluate(map, expr, Boolean.class);
            if (!valid) {
                log.info("[{}][ℹ️️ {}- 断言未通过 ]" + "[{}] {}: {}",
                        taskName,
                        description,
                        expr,
                        uniqueField.getFieldName(),
                        uniqueField.getValue());
                isValid = false;
                break;
            }
        }
        return isValid;
    }

    private void putToCache(TableTemplate template, String md5Digest) {
        // 必须在方法出栈前，将md5放入缓存，因为Spring Cloud Stream默认会重试三次，如果提前放入缓存了，会导致首次重试被判定为无需处理
        urlMD5Service.put(template.getUniqueField().getValue().toString(), md5Digest);
    }


    private boolean tryDelete(TaskVO taskVO, TableTemplate template, HashMap<String, Object> newItem, SinkParams.CheckDelete checkDeleteConfig, ProcessReason reason) {
        // 如果不是 是固定值检测删除并且状态是开启，说明无需处理，直接返回
        if (
                !(Status.on.equals(checkDeleteConfig.getStatus()) && DeleteCheckMode.fixedValue.equals(checkDeleteConfig.getMode()))
        ) {
            return false;
        }
        SinkParams.FixedValueConfig fixedValueConfig = checkDeleteConfig.getFixedValueConfig();
        List<String> predicates = fixedValueConfig.getPredicates();
        // 满足断言条件，才能继续处理
        boolean satisfied = this.isSatisfied(taskVO.getTask(), newItem, template, predicates, "删除");
        if (satisfied) {
            return doDelete(taskVO, template, reason);
        }
        return false;
    }

    public boolean doDelete(TaskVO taskVO, TableTemplate template, ProcessReason reason) {
        TableTemplate.Field uniqueField = template.getUniqueField();
        String tableName = template.getTableName();
        Object value = uniqueField.getValue();
        TableTemplate.LogicDeleteField logicDeleteField = template.getLogicDeleteField();
        boolean isLogicDelete = !ObjectUtils.isEmpty(logicDeleteField);
        String deleteSql;
        if (!isLogicDelete) {
            deleteSql = SqlGenerator.generateDeleteSql(tableName, uniqueField.getFieldName(), uniqueField.getValue());
        } else {
            deleteSql = SqlGenerator.generateLogicDeleteUpdateSql(tableName, template);
        }
        int rows = jdbcService.executeSql(template.getDatasourceName(), deleteSql);
        log.info("[{}][\uD83D\uDDD1️已{}删除{}条][{} {}] " + "{}: {}",
                taskVO.getTask().getName(),
                isLogicDelete ? "逻辑" : "",
                rows,
                reason.getEmoji(),
                reason.name(),
                uniqueField.getFieldName(),
                value);
        if (ProcessReason.external_condition.equals(reason)) {
            return rows > 0;
        }
        if (rows != 1) {
            throw new RuntimeException("期望删除一条，实际删除" + rows + "条数据，请尽快处理");
        }
        return true;
    }


    private boolean isTemplateValid(String taskName, TableTemplate template) {
        boolean isValid = true;
        List<TableTemplate.Field> allField = new ArrayList<>();
        allField.addAll(template.getOtherFields());
        allField.addAll(template.getAutoValueFields());
        allField.add(template.getUniqueField());
        allField.add(template.getCreateTimeField());
        allField.add(template.getUpdateTimeField());
        allField.add(template.getLogicDeleteField());
        allField.removeIf(field -> ObjectUtils.isEmpty(field) || ObjectUtils.isEmpty(field.getRules()));
        for (TableTemplate.Field field : allField) {
            List<String> rules = field.getRules();
            for (String rule : rules) {
                boolean valid = (Boolean) evaluator.evaluate(field.getValue(), rule, Boolean.class);
                if (!valid) {
                    log.warn("[{}][❌️ 模板规则校验未通过]" + "[{}] {}: {}",
                            taskName,
                            rule,
                            field.getFieldName(),
                            field.getValue());
                    isValid = false;
                    break;
                }
            }
        }
        return isValid;
    }


    private void validate(HashMap<String, Object> expressionResultMap, TableTemplate tableTemplate) {
        if (!expressionResultMap.containsKey(tableTemplate.getUniqueField().getFieldName())) {
            throw new RuntimeException("uniqueField [" + tableTemplate.getUniqueField().getFieldName() + "] not found");
        }
    }

    private void mergeValueToTemplate(TaskVO taskVO, TableTemplate template, HashMap<String, Object> expressionResultMap) {
        // 唯一字段一定要使用计算结果
        template.getUniqueField().setValue(expressionResultMap.get(template.getUniqueField().getFieldName()));

        // 将计算结果，填充/覆盖模板中的字段
        for (TableTemplate.Field field : template.getOtherFields()) {
            String fieldName = field.getFieldName();
            if (expressionResultMap.containsKey(fieldName)) {
                Object value = expressionResultMap.get(fieldName);
                field.setValue(value);
            }
        }
    }


    private ProcessReason getReason(TaskVO taskVO, TableTemplate template, String md5) {

        String url = template.getUniqueField().getValue().toString();
        ResourceStatus resourceStatus = urlMD5Service.getResourceStatus(url, md5);

        ProcessMode processMode = taskVO.getTask().getProcessMode();
        if (ResourceStatus.UNCHANGED == resourceStatus && ProcessMode.NORMAL == processMode) {
            log.debug("@@ 数据未更新，无需继续处理");
            return ProcessReason.unchanged;
        } else {
            if (ProcessMode.DEBUG == processMode && ResourceStatus.UNCHANGED == resourceStatus) {
                log.debug("@@ 数据未更新，但当前为调试模式，需继续进行处理");
                return ProcessReason.debug_mode;
            } else if (ResourceStatus.UNKNOWN == resourceStatus) {
                log.debug("@@ 数据状态未知，需进行处理");
                return ProcessReason.cache_not_exist;
            } else if (ResourceStatus.UPDATED == resourceStatus) {
                log.debug("@@ 数据相比缓存中上次记录已更新，需进行处理");
                return ProcessReason.resource_changed;
            }
        }
        return ProcessReason.other;
    }

    private boolean itemExist(SinkParams sinkParams, TableTemplate tableTemplate) {
        String selectExist = SqlGenerator.generateExistedSql(tableTemplate);
        return jdbcService.isExist(tableTemplate.getDatasourceName(), selectExist);
    }

    private Map<String, Object> getExistedItem(TableTemplate template) {
        String datasourceName = template.getDatasourceName();
        return jdbcService.getOne(datasourceName, SqlGenerator.generateFetchSql(template));
    }


    private void tryUpdate(TaskVO taskVO, Map<String, Object> dbItem, HashMap<String, Object> resultMap, String md5Digest, TableTemplate tableTemplate, ProcessReason reason) {
        SinkParams sinkParams = taskVO.getTask().getSinkParams();
        String taskName = taskVO.getTask().getName();
        boolean updateConfigEnabled = sinkParams.getCheckUpdate().getStatus().equals(Status.on);
        if (updateConfigEnabled) {
            String dataSourceName = tableTemplate.getDatasourceName();
            Assert.notNull(dataSourceName, "dataSourceName is null");
            SinkParams.CheckUpdate checkUpdateConfig = sinkParams.getCheckUpdate();
            String fieldName = checkUpdateConfig.getFieldName();
            Object dbValue = dbItem.get(fieldName);
            if (ObjectUtils.isEmpty(dbValue)) {
                log.warn("根据{}来判断数据是否需要更新，但是数据库不存在该字段", fieldName);
                return;
            }
            Object nowValue = resultMap.get(fieldName);
            if (!ObjectUtils.isEmpty(checkUpdateConfig.getJsonPath())) {
                dbValue = cf.jsonExtract(dbValue.toString(), checkUpdateConfig.getJsonPath());
                nowValue = cf.jsonExtract(nowValue.toString(), checkUpdateConfig.getJsonPath());
            } else {
                throw new RuntimeException("根据 [" + fieldName + "] 来判断数据是否需要更新，但是没有配置jsonPath");
            }
            if (dbValue.equals(nowValue)) {
                log.info("[{}][✅ 已是最新][{} {}] " + "{}: {}",
                        taskName,
                        reason.getEmoji(),
                        reason.name(),
                        tableTemplate.getUniqueField().getFieldName(),
                        tableTemplate.getUniqueField().getValue());
                urlMD5Service.put(tableTemplate.getUniqueField().getValue().toString(), md5Digest);
                return;
            } else {
                String update = SqlGenerator.generateUpdateSql(tableTemplate);
                jdbcService.update(dataSourceName, update);
                log.info("[{}][\uD83D\uDD04 已更新][{} {}] " + "{}: {}",
                        taskName,
                        reason.getEmoji(),
                        reason.name(),
                        tableTemplate.getUniqueField().getFieldName(),
                        tableTemplate.getUniqueField().getValue());
            }
        } else {
            log.info("[{}][⏭️ 不检查更新][{} {}] " + "{}: {}",
                    taskName,
                    reason.getEmoji(),
                    reason.name(),
                    tableTemplate.getUniqueField().getFieldName(),
                    tableTemplate.getUniqueField().getValue());
        }
    }


    private void tryInsert(String message, TaskVO taskVO, Map<String, Object> newItem, TableTemplate template, ProcessReason reason) {
        Task task = taskVO.getTask();
        // 执行前置检查
        if (!this.isSatisfied(task, newItem, template, task.getSinkParams().getInsertConfig().getPredicates(), "保存")) {
            return;
        } else {
            jdbcService.update(template.getDatasourceName(), SqlGenerator.generateInsertSql(template));
            log.info("[{}][\uD83D\uDCBE 已保存][{} {}] " + "{}: {}",
                    task.getName(),
                    reason.getEmoji(),
                    reason.name(),
                    template.getUniqueField().getFieldName(),
                    template.getUniqueField().getValue());
        }

        // 执行后置动作
        List<String> postActions = task.getSinkParams().getInsertConfig().getPostActions();
        if (ObjectUtils.isEmpty(postActions)) {
            return;
        }
        for (String expression : postActions) {
            evaluator.evaluate(message, expression, Void.class);
        }
    }

}