package work.xuye.sink.handler;

import work.xuye.common.db.entity.vo.TableTemplate;
import work.xuye.common.enums.DbFieldType;

import java.util.ArrayList;
import java.util.List;

/**
 * @author xuye
 * @since 2023/3/4 20:01
 **/
public class SqlGenerator {

    public static String generateExistedSql(TableTemplate template) {
        StringBuilder sqlBuilder = new StringBuilder();
        sqlBuilder.append("SELECT COUNT(*) FROM `").append(template.getTableName()).append("` WHERE ");

        TableTemplate.Field uniqueField = template.getUniqueField();
        if (uniqueField == null) {
            sqlBuilder.append("1=1");
        } else {
            sqlBuilder.append('`').append(uniqueField.getFieldName()).append('`')
                    .append('=').append(formatFieldValue(uniqueField));
        }

        return sqlBuilder.toString();
    }


    public static String generateUpdateSql(TableTemplate template) {
        StringBuilder sqlBuilder = new StringBuilder();
        sqlBuilder.append("UPDATE `").append(template.getTableName()).append("` SET ");

        // Add field names and values to the query, except uniqueField
        List<TableTemplate.Field> fields = template.getOtherFields();
        boolean isFirstField = true;

        for (TableTemplate.Field field : fields) {
            if (!isFirstField) {
                sqlBuilder.append(", ");
            } else {
                isFirstField = false;
            }
            sqlBuilder.append('`').append(field.getFieldName()).append('`')
                    .append('=').append(formatFieldValue(field));
        }

        TableTemplate.Field updateTimeField = template.getUpdateTimeField();
        if (updateTimeField != null) {
            sqlBuilder.append(", `").append(updateTimeField.getFieldName()).append("`=")
                    .append(generateCurrentTimeFunctionString(updateTimeField.getFieldType()));
        }
        
        sqlBuilder.append(" WHERE ")
                .append('`').append(template.getUniqueField().getFieldName()).append('`')
                .append('=').append(formatFieldValue(template.getUniqueField()));

        return sqlBuilder.toString();
    }


    public static String generateInsertSql(TableTemplate template) {
        StringBuilder sqlBuilder = new StringBuilder();
        sqlBuilder.append("INSERT INTO `").append(template.getTableName()).append("` (");


        List<TableTemplate.Field> otherFields = template.getOtherFields();

        boolean isFirstField = true;

        TableTemplate.Field uniqueField = template.getUniqueField(); // move uniqueField up
        if (uniqueField != null) {
            sqlBuilder.append('`').append(uniqueField.getFieldName()).append('`');
            isFirstField = false;
        }

        for (TableTemplate.Field field : otherFields) {
            if (field == uniqueField) { // skip uniqueField
                continue;
            }
            if (!isFirstField) {
                sqlBuilder.append(", ");
            } else {
                isFirstField = false;
            }
            sqlBuilder.append('`').append(field.getFieldName()).append('`');
        }

        TableTemplate.Field createTimeField = template.getCreateTimeField();
        if (createTimeField != null) {
            sqlBuilder.append(", `").append(createTimeField.getFieldName()).append('`');
        }
        TableTemplate.Field updateTimeField = template.getUpdateTimeField();

        if (updateTimeField != null) {
            sqlBuilder.append(", `").append(updateTimeField.getFieldName()).append('`');
        }

        sqlBuilder.append(") VALUES (");

        isFirstField = true;

        if (uniqueField != null) { // move uniqueField value up
            sqlBuilder.append(formatFieldValue(uniqueField));
            isFirstField = false;
        }

        for (TableTemplate.Field field : otherFields) {
            if (field == uniqueField) { // skip uniqueField
                continue;
            }
            if (!isFirstField) {
                sqlBuilder.append(", ");
            } else {
                isFirstField = false;
            }
            sqlBuilder.append(formatFieldValue(field));
        }


        if (createTimeField != null) {
            sqlBuilder.append(", ").append(generateCurrentTimeFunctionString(createTimeField.getFieldType()));
        }

        if (updateTimeField != null) {
            sqlBuilder.append(", ").append(generateCurrentTimeFunctionString(updateTimeField.getFieldType()));
        }

        sqlBuilder.append(")");
        return sqlBuilder.toString();
    }


    public static String formatFieldValue(TableTemplate.Field field) {
        Object value = field.getValue();
        if (value == null) {
            return "NULL";
        }
        switch (field.getFieldType()) {
            case json:
            case string:
                String stringValue = value.toString().replace("\\", "\\\\")
                        .replace("'", "\\'");
                return "'" + stringValue + "'";
            default:
                return value.toString();
        }
    }


    public static String generateCurrentTimeFunctionString(DbFieldType fieldType) {
        switch (fieldType) {
            case datetime:
                return "NOW()";
            case timestamp:
                return "CURRENT_TIMESTAMP()";
            case date:
                return "CURDATE()";
            case time:
                return "CURTIME()";
            default:
                return "NULL";
        }
    }

    public static String generateFetchSql(TableTemplate tableTemplate) {
        StringBuilder sb = new StringBuilder();

        sb.append("SELECT * FROM `")
                .append(tableTemplate.getTableName())
                .append("` WHERE ");

        TableTemplate.Field uniqueField = tableTemplate.getUniqueField();

        if (uniqueField == null) {
            throw new IllegalArgumentException("TableTemplate does not have unique field.");
        }

        sb.append('`')
                .append(uniqueField.getFieldName())
                .append("`='")
                .append(escapeValue(uniqueField.getValue().toString()))
                .append("'");

        return sb.toString();
    }

    public static String escapeValue(String value) {
        return value.replace("\\", "\\\\")
                .replace("'", "\\'");
    }

    public static String generateGivenFieldSql(String tableName, String selectField, String condition, String lastValue) {
        return "SELECT `" + selectField + "` FROM `" + tableName + "` WHERE " +
                condition + '\'' + lastValue + '\'';
    }

    public static String generateGivenFieldSql(String tableName, String condition, String lastValue) {
        return "SELECT * FROM `" + tableName + "` WHERE " +
                condition + '\'' + lastValue + '\'';
    }

    public static String generateDeleteSql(String tableName, String field, Object value) {
        return "DELETE FROM " + tableName + " WHERE " + field + "='" + value + "'" + " LIMIT 1;";
    }

    public static List<String> generateDeleteSqls(String tableName, String fieldName, List<String> dbs) {
        List<String> result = new ArrayList<>();
        for (String db : dbs) {
            result.add(generateDeleteSql(tableName, fieldName, db));
        }
        return result;
    }


    public static String generateLogicDeleteUpdateSql(String tableName, TableTemplate template) {

        TableTemplate.Field updateTimeField = template.getUpdateTimeField();
        TableTemplate.LogicDeleteField logicDeleteField = template.getLogicDeleteField();
        String logicDeleteFieldName = logicDeleteField.getFieldName();
        String logicDeleteValue = logicDeleteField.getDeleteValue();
        TableTemplate.Field uniqueField = template.getUniqueField();
        String uniqueFieldFieldName = uniqueField.getFieldName();
        Object uniqueFieldValue = uniqueField.getValue();
        TableTemplate.Field updateField = template.getUpdateField();

        StringBuilder sqlBuilder = new StringBuilder();
        sqlBuilder.append("UPDATE `").append(tableName).append("` SET ");
        sqlBuilder.append('`').append(logicDeleteFieldName).append('`')
                .append('=').append(logicDeleteValue);
        sqlBuilder.append(", `").append(updateField.getFieldName()).append("`=")
                .append(updateField.getValue());
        if (updateTimeField != null) {
            sqlBuilder.append(", `").append(updateTimeField.getFieldName()).append("`=")
                    .append(generateCurrentTimeFunctionString(updateTimeField.getFieldType()));
        }

        sqlBuilder.append(" WHERE ")
                .append('`').append(uniqueFieldFieldName).append('`')
                .append('=').append('\'').append(uniqueFieldValue.toString())
                .append("' LIMIT 1");
        return sqlBuilder.toString();
    }
}
