package work.xuye.sink.yueke;

import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.codec.digest.DigestUtils;
import org.springframework.stereotype.Service;
import org.springframework.util.ObjectUtils;
import work.xuye.common.db.entity.Mapping;
import work.xuye.common.db.entity.Task;
import work.xuye.common.db.entity.Template;
import work.xuye.common.db.entity.vo.TableTemplate;
import work.xuye.common.db.service.TaskManager;
import work.xuye.common.dto.TaskVO;
import work.xuye.common.enums.ProcessMode;
import work.xuye.common.enums.ProcessReason;
import work.xuye.common.properties.YueKeProperties;
import work.xuye.common.store.NsKVMapStore;
import work.xuye.sink.handler.ItemHandler;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

/**
 * @author xuye
 * @since 2023/7/10 10:15
 **/
@Slf4j
@Service
@RequiredArgsConstructor
public class YueKeRevokeService {

    private final YueKeProperties yueKeProperties;
    private final TaskManager taskManager;
    private final NsKVMapStore mapStore;
    private final ItemHandler itemHandler;


    private static String getSign(Long timestamp, String reqId, String secret) {
        String params = timestamp + reqId + secret;
        return DigestUtils.md5Hex(params);
    }

    public void handleRevokeRequest(YueKeRevokeNewsIDsVO ids, YueKeRequestModel requestModel) {
        // 根据沟通，阅客对同一篇资讯会下发两次id，一次是普通的id，一次是带有:的id，这里需要过滤掉带:的id
        if (this.afterClean(ids).isEmpty()) {
            log.info("revoke ids is empty, ignore");
            return;
        }
        log.info("待下架的资讯id列表: {}", ids);
        // 确保是阅客的请求
        this.checkRequest(requestModel);
        log.info("请求合法，开始下架");
        // 执行下架
        this.revoke(ids.getData().getMessageIdList());
    }

    public void revoke(List<String> idList) {
        // 前提条件：阅客的任务名包含 yueke，并且只有一条记录
        Task task = taskManager.getTaskService().lambdaQuery()
                .eq(Task::getName, yueKeProperties.getTaskName())
                .eq(Task::getProcessMode, ProcessMode.NORMAL)
                .select(Task::getId)
                .orderBy(true, true, Task::getCreateTime)
                .last("limit 1")
                .one();
        if (ObjectUtils.isEmpty(task)) {
            log.warn("接收到阅客的下架请求，但是库中没有此任务，跳过处理");
            return;
        }
        TaskVO taskVO = taskManager.getTaskInfoByTaskId(task.getId());

        for (String id : idList) {
            String url = mapStore.getValue(yueKeProperties.getNamespaceOfIdUrlMap(), id);
            if (ObjectUtils.isEmpty(url)) {
                log.warn("接收到阅客的下架请求，但是库中没有此记录，跳过处理");
                continue;
            }
            List<Template> templateList = new ArrayList<>();
            for (String templateName : taskVO.getTemplateMappingMap().values().stream().map(Mapping::getTemplate).collect(Collectors.toList())) {
                Template templateByName = taskManager.getTemplateByName(templateName);
                templateList.add(templateByName);
            }
            for (Template template : templateList) {
                TableTemplate tableTemplate = template.getTableTemplate();
                tableTemplate.getUniqueField().setValue(url);
                if (itemHandler.doDelete(taskVO, tableTemplate, ProcessReason.external_condition)) {
                    log.info("阅客下架成功，资讯id: {}, url: {}", id, url);
                } else {
                    log.info("阅客下架失败，资讯id: {}, url: {}", id, url);
                }
            }
        }
    }

    private List<String> afterClean(YueKeRevokeNewsIDsVO idsVO) {
        List<String> messageIdList = idsVO.getData().getMessageIdList();
        messageIdList.removeAll(
                messageIdList.stream().filter(id -> id.contains(":")).collect(Collectors.toList())
        );
        return messageIdList;
    }

    private void checkRequest(YueKeRequestModel requestModel) {

        try {
            if (ObjectUtils.isEmpty(requestModel)) {
                throw new YueKeException(null, YueKeResEnum.SECURITY_ERROR_11001);
            }
            Long requestTimeStamp = requestModel.getTimestamp();
            String requestId = requestModel.getReqId();
            String requestSign = requestModel.getSign();

            //参数完整性校验
            if (ObjectUtils.isEmpty(requestId) || ObjectUtils.isEmpty(requestSign) || null == requestTimeStamp) {
                throw new YueKeException(requestId, YueKeResEnum.SECURITY_ERROR_11002);
            }
            int requestIdLength = requestId.length();
            if (requestIdLength < YueKeConstants.REQ_ID_MIN_LENGTH) {
                throw new YueKeException(requestId, YueKeResEnum.SECURITY_ERROR_11007);
            } else if (requestIdLength > YueKeConstants.REQ_ID_MAX_LENGTH) {
                throw new YueKeException(requestId, YueKeResEnum.SECURITY_ERROR_11008);
            }
            //时间有效期验证
            Long nowTime = System.currentTimeMillis();
            long minValidTime = nowTime - YueKeConstants.VALID_MS;
            long maxValidTime = nowTime + YueKeConstants.VALID_MS;
            if (!(requestTimeStamp >= minValidTime && requestTimeStamp <= maxValidTime)) {
                throw new YueKeException(requestId, YueKeResEnum.SECURITY_ERROR_11003);
            }
            //sign验证
            String sign = getSign(requestTimeStamp, requestId, yueKeProperties.getSecret());
            if (!requestSign.equals(sign)) {
                throw new YueKeException(requestId, YueKeResEnum.SECURITY_ERROR_11006);
            }
        } catch (YueKeException yueKeException) {
            log.warn("check revoke param error: ", yueKeException);
            throw yueKeException;
        } catch (Exception ex) {
            log.error("other exception : {}", ex.getMessage(), ex);
            throw new YueKeException(null, YueKeResEnum.SECURITY_ERROR_11501);
        }

    }


}
