package work.xuye.source.handler;

import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.cloud.stream.function.StreamBridge;
import org.springframework.expression.Expression;
import org.springframework.expression.spel.standard.SpelExpressionParser;
import org.springframework.expression.spel.support.StandardEvaluationContext;
import org.springframework.messaging.Message;
import org.springframework.messaging.support.MessageBuilder;
import org.springframework.stereotype.Component;
import org.springframework.util.DigestUtils;
import org.springframework.util.ObjectUtils;
import work.xuye.common.constant.*;
import work.xuye.common.db.entity.Task;
import work.xuye.common.db.entity.vo.TaskMiniVO;
import work.xuye.common.db.service.TaskManager;
import work.xuye.common.dto.HttpRequestParams;
import work.xuye.common.dto.HttpRes;
import work.xuye.common.enums.ProcessMode;
import work.xuye.common.enums.RequestMode;
import work.xuye.common.enums.ResourceStatus;
import work.xuye.common.service.UrlMD5Service;
import work.xuye.common.utils.DebugUtil;
import work.xuye.common.utils.ID;
import work.xuye.source.request.RequestClient;
import work.xuye.source.spel.CustomRequest;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;

/**
 * @author xuye
 * @since 2023/2/19 16:37
 **/
@Slf4j
@Component
@RequiredArgsConstructor
public class SourceHandler {

    public final CustomRequest cr;
    private final SpelExpressionParser parser;
    private final RequestClient requestClient;
    private final UrlMD5Service urlMD5Service;
    private final StreamBridge streamBridge;
    private final TaskManager taskManager;

    public List<Message<HttpRes>> handle(Message<TaskMiniVO> message) {
        TaskMiniVO taskMiniVO = message.getPayload();
        if (DebugUtil.isSkip(taskMiniVO)) {
            return null;
        }
        Task task = taskManager.getTaskInfoByTaskId(taskMiniVO.getId()).getTask();
        if (DebugUtil.isSkip(task)) {
            return null;
        }
        HttpRequestParams req = task.getRequestParams();
        String taskName = task.getName();
        HttpRes res = null;
        List<HttpRes> result = null;
        if (RequestMode.normal.equals(req.getMode())) {
            res = this.request(req);
            result = this.handleRes(task, res);
        } else if (RequestMode.SpEL.equals(req.getMode())) {

            StandardEvaluationContext context = new StandardEvaluationContext();
            Expression expression = parser.parseExpression(task.getRequestParams().getSpELConfig().getExpression());
            List<HttpRes> value = expression.getValue(context, this, List.class);
            result = this.handleRes(task, value);
        }
        if (ObjectUtils.isEmpty(result)) {
            return null;
        }
        if (ProcessMode.WATCH.equals(task.getProcessMode())) {
            for (HttpRes r : result) {
                Message<String> watchMessage = MessageBuilder
                        .withPayload(r.getBody())
                        .setHeader(MessageConstants.CURRENT_STAGE, StageConstants.SOURCE_WATCH)
                        .setHeader(MessageConstants.TASK_TRACE_ID, message.getHeaders().get(MessageConstants.TASK_TRACE_ID))
                        .setHeader(MessageConstants.SOURCE_TRACE_ID, ID.generate())
                        .setHeader(MessageConstants.TASK_ID, message.getHeaders().get(MessageConstants.TASK_ID))
                        .setHeader(MessageConstants.TASK_NAME, message.getHeaders().get(MessageConstants.TASK_NAME))
                        .setHeader(SnapshotConstants.URL, req.getUrl())
                        .setHeader(SnapshotConstants.RESOURCE_STATUS, r.getResourceStatus())
                        .setHeader(SnapshotConstants.STATUS, r.getStatus())
                        .build();

                boolean send = streamBridge.send(taskName + BindingConstants.SNAPSHOT_OUT_POSTFIX, watchMessage);
                if (!send) {
                    throw new RuntimeException("send message failed");
                }
            }
            return null;
        }

        ArrayList<Message<HttpRes>> messageList = new ArrayList<>();
        result.forEach(
                item -> {
                    messageList.add(MessageBuilder
                            .withPayload(item)
                            .setHeader(MessageConstants.CURRENT_STAGE, StageConstants.SOURCE)
                            .setHeader(MessageConstants.TASK_TRACE_ID, message.getHeaders().get(MessageConstants.TASK_TRACE_ID))
                            .setHeader(MessageConstants.SOURCE_TRACE_ID, ID.generate())
                            .setHeader(MessageConstants.TASK_ID, message.getHeaders().get(MessageConstants.TASK_ID))
                            .setHeader(MessageConstants.TASK_NAME, message.getHeaders().get(MessageConstants.TASK_NAME))
                            .setHeader(MessageConstants.SEED_URL, item.getRequestParams().getUrl())
                            .build());
                }
        );
        return messageList;


    }


    private void setResourceStatus(HttpRes res) {
        String md5Digest = DigestUtils.md5DigestAsHex(res.getBody().getBytes());
        ResourceStatus resourceStatus = urlMD5Service.getResourceStatus(res.getRequestParams().getUrl(), md5Digest);
        urlMD5Service.put(res.getRequestParams().getUrl(), md5Digest);
        res.setResourceStatus(resourceStatus);
    }


    private List<HttpRes> handleRes(Task task, List<HttpRes> resList) {
        if (ObjectUtils.isEmpty(resList)) {
            return null;
        }
        List<HttpRes> out = new ArrayList<>();
        for (HttpRes res : resList) {
            if (!res.getStatus().is2xxSuccessful()) {
                log.warn("request failed, url: {}, status: {}", res.getRequestParams().getUrl(), res.getStatus());
                continue;
            }
            this.setResourceStatus(res);
            this.log(res, task);
            if (res.getResourceStatus() != ResourceStatus.UNCHANGED || ProcessMode.DEBUG == task.getProcessMode()) {
                out.add(res);
            }
        }
        return out;

    }

    private List<HttpRes> handleRes(Task task, HttpRes res) {
        List<HttpRes> out = new ArrayList<>();
        if (!res.getStatus().is2xxSuccessful()) {
            log.info("log for debug：{}", res);
            log.warn("request failed, url: {}, status: {}", res.getRequestParams().getUrl(), res.getStatus());
            return null;
        }
        this.setResourceStatus(res);
        this.log(res, task);
        if (res.getResourceStatus().equals(ResourceStatus.UNCHANGED) && !ProcessMode.DEBUG.equals(task.getProcessMode())) {
            return null;
        } else {
            out.add(res);
            return out;
        }
    }


    private void log(HttpRes res, Task task) {
        ProcessMode processMode = task.getProcessMode();
        HttpRequestParams req = res.getRequestParams();
        String emoji = "";
        if (processMode.equals(ProcessMode.NORMAL)) {
            emoji = "";
        } else if (processMode.equals(ProcessMode.DEBUG)) {
            emoji = "🤖";
        } else if (processMode.equals(ProcessMode.WATCH)) {
            emoji = "👀";
        }
        if (res.getStatus().is2xxSuccessful()) {
            log.info("[{}-{}] [{}{} {}] [{} {} {}ms] {} ", task.getId(), task.getName(),
                    res.getResourceStatus().getEmoji(), emoji,
                    res.getResourceStatus().getName(), req.getMethod(), res.getStatus(), res.getCostTimeMillis(), req.getUrl());
        } else {
            log.warn("@@ method: [{}], status: [{}], url: [{}], request: [{}], response: [{}]",
                    req.getMethod(), res.getStatus(), req.getUrl(), req.getBody(), res.getBody());
        }
    }

    public HttpRes request(HttpRequestParams req) {
        this.preHandleUrl(req);
        this.preHandleBody(req);
        return requestClient.execute(req);
    }

    private void preHandleBody(HttpRequestParams request) {
        Map<String, Object> body = request.getBody();
        if (ObjectUtils.isEmpty(body)) {
            return;
        }
        for (String key : body.keySet()) {
            Object value = body.get(key);
            Matcher matcher = RegexConstants.PLACEHOLDER_PATTERN.matcher(value.toString());
            if (matcher.find()) {
                String placeholder = matcher.group();
                String expr = request.getPlaceholderExpressions().get(placeholder);
                Object result = parser.parseExpression(expr).getValue();
                body.put(key, result);
            }
        }
        request.setBody(body);
    }

    private void preHandleUrl(HttpRequestParams request) {
        String url = request.getUrl();
        Matcher matcher = RegexConstants.PLACEHOLDER_PATTERN.matcher(url);
        Map<String, Object> patternMap = new HashMap<>();
        while (matcher.find()) {
            String placeholder = matcher.group();
            patternMap.put(placeholder, null);
        }

        if (ObjectUtils.isEmpty(patternMap)) {
            return;
        }
        Map<String, String> placeholderExpressions = request.getPlaceholderExpressions();

        for (String key : patternMap.keySet()) {
            String expr = placeholderExpressions.get(key);
            Object value = parser.parseExpression(expr).getValue();
            patternMap.put(key, value);
        }
        for (String key : patternMap.keySet()) {
            url = url.replace(key, patternMap.get(key).toString());
        }
        request.setUrl(url);
    }


}



