package work.xuye.source.request;

import com.google.gson.Gson;
import lombok.extern.slf4j.Slf4j;
import okhttp3.*;
import okio.BufferedSource;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Primary;
import org.springframework.http.HttpMethod;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Component;
import org.springframework.util.ObjectUtils;
import org.springframework.util.StopWatch;
import work.xuye.common.dto.HttpRequestParams;
import work.xuye.common.dto.HttpRes;
import work.xuye.source.properties.SourceProperties;
import work.xuye.source.util.CharsetUtil;

import java.net.SocketTimeoutException;
import java.nio.charset.Charset;
import java.util.Map;
import java.util.concurrent.TimeUnit;

/**
 * @author xuye
 * @since 2023/4/7 16:29
 **/
@Slf4j
@Primary
@Component
public class OkHttpRequestClient implements RequestClient {

    private final Gson gson;
    private final OkHttpClient client;

    @Autowired
    public OkHttpRequestClient(Gson gson, SourceProperties sourceProperties) {
        this.gson = gson;
        SourceProperties.Pool poolConfig = sourceProperties.getPool();
        this.client = new OkHttpClient.Builder()
                .connectTimeout(sourceProperties.getTimeoutSeconds(), TimeUnit.SECONDS)
                .connectionPool(new ConnectionPool(
                        poolConfig.getMaxIdleConnections(),
                        poolConfig.getKeepAliveSeconds(),
                        TimeUnit.SECONDS))
                .followRedirects(true)
                .build();
        log.info("init OkHttpRequestClient success, timeoutSeconds: {}, maxIdleConnections: {}, keepAliveSeconds: {}",
                sourceProperties.getTimeoutSeconds(),
                poolConfig.getMaxIdleConnections(),
                poolConfig.getKeepAliveSeconds());
    }


    @Override
    public HttpRes execute(HttpRequestParams request) {
        StopWatch stopWatch = new StopWatch();
        stopWatch.start();
        Request.Builder requestBuilder = new Request.Builder();
        // request url
        requestBuilder.url(request.getUrl());

        // request header
        if (!ObjectUtils.isEmpty(request.getHeaders())) {
            requestBuilder.headers(Headers.Companion.of(request.getHeaders()));
        }

        // request body
        Map<String, Object> body = request.getBody();
        String type = request.getMediaType();
        RequestBody requestBody = null;
        //如果不是GET请求，并且请求体不为空才能构建请求体
        if (!request.getMethod().equalsIgnoreCase(HttpMethod.GET.name())
                && !ObjectUtils.isEmpty(request.getBody())
                && !ObjectUtils.isEmpty(request.getBody().keySet())) {
            // form url
            if (org.springframework.http.MediaType.APPLICATION_FORM_URLENCODED_VALUE.equals(type)) {
                FormBody.Builder formBodyBuilder = new FormBody.Builder();
                for (Map.Entry<String, Object> entry : request.getBody().entrySet()) {
                    formBodyBuilder.add(entry.getKey(), entry.getValue().toString());
                }
                requestBody = formBodyBuilder.build();
            } else if (org.springframework.http.MediaType.APPLICATION_JSON_VALUE.equals(type)) {
                MediaType mediaType = MediaType.Companion.parse(org.springframework.http.MediaType.APPLICATION_JSON_VALUE);
                requestBody = RequestBody.Companion.create(gson.toJson(body), mediaType);
            } else {
                throw new RuntimeException("暂不支持的请求类型");
            }
        }

        //如果非get请求，并且请求体仍然为空，需要构建一个空的请求体，因为okhttp不支持null请求体
        if (!request.getMethod().equalsIgnoreCase(HttpMethod.GET.name()) && ObjectUtils.isEmpty(requestBody)) {
            requestBody = new FormBody.Builder().build();
        }
        requestBuilder.method(request.getMethod(), requestBody);


        // res
        HttpRes httpRes = new HttpRes();
        Request httpRequest = requestBuilder.build();
        try (Response response = client.newCall(httpRequest).execute()) {
            httpRes.requestParams(request).headers(response.headers().toMultimap()).status(HttpStatus.valueOf(response.code()));
            if (!ObjectUtils.isEmpty(response.body())) {
                BufferedSource source = response.body().source();
                Charset charset = CharsetUtil.getCharsetByName(request.getCharset());
                String responseBody = source.readByteString().string(charset);
                httpRes.body(responseBody);
                source.close();
            }
        } catch (SocketTimeoutException e) {
            httpRes.status(HttpStatus.REQUEST_TIMEOUT);
        } catch (Exception e) {
            log.error("execute request error+", e);
            httpRes.status(HttpStatus.INTERNAL_SERVER_ERROR);
        }
        stopWatch.stop();
        return httpRes.costTimeMillis(stopWatch.getTotalTimeMillis());

    }

}
