package work.xuye.source.spel;

import com.google.gson.JsonObject;
import com.google.gson.JsonParser;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import net.minidev.json.JSONObject;
import org.apache.commons.codec.digest.DigestUtils;
import org.springframework.stereotype.Component;
import org.springframework.web.util.UriComponentsBuilder;
import work.xuye.common.dto.HttpRequestParams;
import work.xuye.common.dto.HttpRes;
import work.xuye.common.properties.YueKeProperties;
import work.xuye.common.store.NsKVMapStore;
import work.xuye.source.request.RequestClient;

import java.util.*;

/**
 * @author xuye
 * @since 2023/7/7 16:29
 **/
@Slf4j
@Component
@RequiredArgsConstructor
public class YueKeRequest {


    private final YueKeProperties yueKeProperties;
    private final RequestClient requestClient;
    private final NsKVMapStore mapStore;

    private final Map<Integer, String> resCodeMap = new HashMap<>();

    {
        resCodeMap.put(50000, "error");
        resCodeMap.put(11001, "metadata is empty");
        resCodeMap.put(11002, "some params in metadata is empty");
        resCodeMap.put(11003, "timestamp valid time is 3 min");
        resCodeMap.put(11004, "reqId is used");
        resCodeMap.put(11006, "sign is not valid");
        resCodeMap.put(11007, "reqId is too short, min length is 10");
        resCodeMap.put(11008, "reqId is too long, max length is 50");
        resCodeMap.put(11501, "other error");
    }

    public List<HttpRes> run(String channelUrlPrefix, String listPageUrlPrefix, String env, String appId, String appSecret, boolean dup) {
        List<HttpRes> resList = new ArrayList<>();
        String channelUrl = UriComponentsBuilder.
                fromUriString(channelUrlPrefix)
                .queryParam("appId", appId).toUriString();
        HashMap<String, String> channelMap = new HashMap<>();
        JsonParser.parseString(this.get(channelUrl, appId, appSecret).getBody())
                .getAsJsonObject().get("data").getAsJsonArray()
                .forEach(
                        item -> {
                            String channelId = item.getAsJsonObject().get("id").getAsString();
                            String channelName = item.getAsJsonObject().get("name").getAsString();
                            channelMap.put(channelId, channelName);
                            String value = mapStore.getValue(yueKeProperties.getNamespaceOfChannelIdNameMap(), channelId);
                            if (!channelName.equals(value)) {
                                mapStore.upsert(yueKeProperties.getNamespaceOfChannelIdNameMap(), channelId, channelName);
                            }
                        }
                );
        log.info("获取频道列表成功，频道列表：{}", channelMap);

        for (String channelId : channelMap.keySet()) {

            String url = UriComponentsBuilder.fromUriString(listPageUrlPrefix)
                    .queryParam("appId", appId)
                    .queryParam("channelId", channelId)
                    .queryParam("userId", env)
                    .queryParam("pageSize", 20)
                    .queryParam("dup", dup)
                    .toUriString();
            HttpRes httpRes = this.get(url, appId, appSecret);
            httpRes.getRequestParams().setUrl(url);
            resList.add(httpRes);
            log.info("获取频道[{}]下的资讯成功", channelMap.get(channelId));
        }
        return resList;
    }

    private HttpRes get(String url, String appId, String appSecret) {
        String authKey = this.getAuthKey(appId, appSecret);
        HttpRequestParams requestParams = new HttpRequestParams();
        requestParams.setUrl(url);
        requestParams.setHeaders(Map.of("metadata", authKey));
        HttpRes httpRes = requestClient.execute(requestParams);
        JsonObject object = JsonParser.parseString(httpRes.getBody()).getAsJsonObject();
        int code = object.get("code").getAsInt();
        if (code != 10000) {
            throw new RuntimeException("请求，失败原因：" + resCodeMap.get(code));
        }
        return httpRes;
    }

    private String getAuthKey(String appId, String appSecret) {
        Long timestamp_req = System.currentTimeMillis();
        String reqId_req = UUID.randomUUID().toString().replaceAll("-", "");
        String unencryptedParams = appId + timestamp_req + reqId_req + appSecret;
        String sign = DigestUtils.md5Hex(unencryptedParams);
        JSONObject jsonObj = new JSONObject();
        jsonObj.put("appId", appId);
        jsonObj.put("reqId", reqId_req);
        jsonObj.put("timestamp", timestamp_req);
        jsonObj.put("sign", sign);
        return jsonObj.toJSONString();
    }
}
