package work.xuye.transformer.handler;

import com.google.gson.JsonArray;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.cloud.stream.function.StreamBridge;
import org.springframework.messaging.Message;
import org.springframework.messaging.support.MessageBuilder;
import org.springframework.stereotype.Component;
import org.springframework.util.Assert;
import org.springframework.util.ObjectUtils;
import work.xuye.common.constant.BindingConstants;
import work.xuye.common.constant.MessageConstants;
import work.xuye.common.constant.StageConstants;
import work.xuye.common.db.entity.Task;
import work.xuye.common.db.entity.vo.ParseParams;
import work.xuye.common.db.entity.vo.SinkParams;
import work.xuye.common.db.entity.vo.TransformParams;
import work.xuye.common.db.service.TaskManager;
import work.xuye.common.dto.HttpRes;
import work.xuye.common.dto.TaskVO;
import work.xuye.common.enums.*;
import work.xuye.common.service.UrlMD5Service;
import work.xuye.common.utils.DebugUtil;
import work.xuye.common.utils.ID;
import work.xuye.common.utils.JsonPathUtil;
import work.xuye.transformer.transformer.MessageTransformer;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * @author xuye
 * @since 2023/2/19 17:04
 **/
@Slf4j
@Component
@RequiredArgsConstructor
public class TransformHandler {
    private final StreamBridge streamBridge;


    private final Map<String, MessageTransformer> transformMessageMap;

    private final TaskManager taskManager;
    private final UrlMD5Service urlMD5Service;


    public List<Message<String>> handle(Message<HttpRes> message) {
        try {
            return this.doHandle(message);
        } catch (Exception e) {
            urlMD5Service.removeUrlMD5(this.getSeedUrl(message));
            throw new RuntimeException(e);
        }
    }


    public List<Message<String>> doHandle(Message<HttpRes> message) {
        // 获取任务信息
        TaskVO taskVO = taskManager.getTaskInfoByTaskId((Integer) message.getHeaders().get(MessageConstants.TASK_ID));
        if (DebugUtil.isSkip(taskVO)) {
            return null;
        }
        ProcessReason processReason = processReason(message, taskVO);
        // 如果资源状态是未变化，直接结束执行
        if (ProcessReason.unchanged.equals(processReason) || ProcessReason.end.equals(processReason)) {
            // todo 考虑是否有必要加日志
            return null;
        }

        // 将结果根据配置的transformer进行转换
        this.transformResult(message, taskVO);

        if (ObjectUtils.isEmpty(message.getPayload().getBody())) {
            log.warn("@@ 转换后的结果为空，不继续处理");
            return null;
        }

        // 消失模式处理
        this.handleDisappear(message, taskVO);

        // 根据jsonpath拆分后下发
        return this.splitAndEnqueue(message, taskVO, processReason);
    }

    private ProcessReason processReason(Message<HttpRes> message, TaskVO taskVO) {
        HttpRes httpRes = message.getPayload();
        ProcessMode processMode = taskVO.getTask().getProcessMode();
        if (ProcessMode.WATCH.equals(processMode)) {
            log.debug("@@ 此任务为监听模式，不会进行任何处理");
            return ProcessReason.end;
        }

        // 理论上不会出现这种情况，因为符合这个条件的消息source是不会下发的，如果出现了，说明是消息被人为重发了
        if (ResourceStatus.UNCHANGED.equals(httpRes.getResourceStatus()) && ProcessMode.NORMAL.equals(processMode)) {
            log.warn("@@ 经md5判断，此URL相对上次未更新，且为正常处理模式，无需做进一步处理");
            return ProcessReason.unchanged;
        } else if (ProcessMode.DEBUG.equals(processMode) && ResourceStatus.UNCHANGED.equals(httpRes.getResourceStatus())) {
            log.debug("@@ 经md5判断，此URL相对上次未更新，但处于调试模式，需继续进行处理");
            return ProcessReason.debug_mode;
        } else if (ResourceStatus.UNKNOWN.equals(httpRes.getResourceStatus())) {
            log.debug("@@ 经md5判断，此URL为首次处理，需进行处理");
            return ProcessReason.cache_not_exist;
        } else if (ResourceStatus.UPDATED.equals(httpRes.getResourceStatus())) {
            log.debug("@@ 经md5判断，此URL相对上次更新了，需进行处理");
            return ProcessReason.resource_changed;
        }

        return ProcessReason.other;
    }


    private void transformResult(Message<HttpRes> message, TaskVO taskVO) {

        String result = message.getPayload().getBody();
        Task task = taskVO.getTask();
        TransformParams transformParams = task.getTransformParams();
        if (ObjectUtils.isEmpty(transformParams)) {
            if (task.getProcessMode().getMode().equals(ProcessMode.WATCH.getMode())) {
                return;
            } else {
                throw new RuntimeException(task.getName() + " transformParams is null");
            }
        }
        List<String> handlers = transformParams.getHandlers();
        Assert.notNull(handlers, "handlers is null");
        String seedUrl = this.getSeedUrl(message);

        for (String handler : handlers) {
            MessageTransformer messageTransformer = transformMessageMap.get(handler);
            if (messageTransformer == null) {
                log.warn("handler {} not found", handler);
                continue;
            }
            result = messageTransformer.transform(result, seedUrl);
        }
        message.getPayload().setBody(result);

    }


    private void handleDisappear(Message<HttpRes> message, TaskVO taskVO) {
        String seedUrl = this.getSeedUrl(message);
        SinkParams.CheckDelete checkDelete = taskVO.getTask().getSinkParams().getCheckDelete();
        ParseParams parseParams = taskVO.getTask().getParseParams();
        String result = message.getPayload().getBody();

        if (checkDelete.getStatus().equals(Status.on) && checkDelete.getMode().equals(DeleteCheckMode.disappear)) {
            JsonArray elements = JsonPathUtil.extractAsJsonArray(result, parseParams.getItemsUniqueKeyPath(), seedUrl);
            if (!ObjectUtils.isEmpty(elements) && elements.size() > 0) {
                Message<String> keysMsg = MessageBuilder.withPayload(elements.toString())
                        .setHeader(MessageConstants.SEED_URL, message.getHeaders().get(MessageConstants.SEED_URL))
                        .setHeader(MessageConstants.TASK_ID, message.getHeaders().get(MessageConstants.TASK_ID))
                        .setHeader(MessageConstants.TASK_NAME, message.getHeaders().get(MessageConstants.TASK_NAME))
                        .setHeader(MessageConstants.TASK_TRACE_ID, message.getHeaders().get(MessageConstants.TASK_TRACE_ID))
                        .setHeader(MessageConstants.SOURCE_TRACE_ID, message.getHeaders().get(MessageConstants.SOURCE_TRACE_ID))
                        .setHeader(MessageConstants.CURRENT_STAGE, StageConstants.TRANSFORMER)
                        .setHeader(MessageConstants.TRANSFORMER_TRACE_ID, ID.generate())
                        .build();
                boolean send = streamBridge.send(BindingConstants.TRANSFORMER_KEYS_OUT, keysMsg);
                if (!send) {
                    throw new RuntimeException("send message failed");
                }
            }

        }

    }


    private ArrayList<Message<String>> splitAndEnqueue(Message<HttpRes> message, TaskVO taskVO, ProcessReason processReason) {

        ParseParams parseParams = taskVO.getTask().getParseParams();

        ArrayList<Message<String>> results = new ArrayList<>();
        String transformerTraceId = ID.generate();
        String seedUrl = this.getSeedUrl(message);
        String result = message.getPayload().getBody();


        if (!ObjectUtils.isEmpty(parseParams) && !ObjectUtils.isEmpty(parseParams.getItemsPath())) {

            JsonArray array = JsonPathUtil.extractAsJsonArray(result, parseParams.getItemsPath(), seedUrl);
            array.forEach(item -> {
                Message<String> itemMessage = MessageBuilder.withPayload(item.toString())
                        .setHeader(MessageConstants.CURRENT_STAGE, StageConstants.TRANSFORMER)
                        .setHeader(MessageConstants.TASK_TRACE_ID, message.getHeaders().get(MessageConstants.TASK_TRACE_ID))
                        .setHeader(MessageConstants.SOURCE_TRACE_ID, message.getHeaders().get(MessageConstants.SOURCE_TRACE_ID))
                        .setHeader(MessageConstants.TRANSFORMER_TRACE_ID, transformerTraceId)
                        .setHeader(MessageConstants.TASK_ID, message.getHeaders().get(MessageConstants.TASK_ID))
                        .setHeader(MessageConstants.TASK_NAME, message.getHeaders().get(MessageConstants.TASK_NAME))
                        .setHeader(MessageConstants.SEED_URL, message.getHeaders().get(MessageConstants.SEED_URL))

                        .build();
                results.add(itemMessage);
            });
            log.info("@ [{}][{}{}] 下发转换后的消息，数量：{}", taskVO.getTask().getName(), processReason.getEmoji(), processReason, results.size());
        } else {
            log.warn("@@ 任务[{}]没有配置itemsPath", taskVO.getTask().getName());
        }
        return results;
    }

    private String getSeedUrl(Message<HttpRes> message) {
        return (String) message.getHeaders().get(MessageConstants.SEED_URL);
    }


}
